# !/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Oct 31 16:23:13 2018

@author: samschott
"""
__version__ = "0.4.0"
__author__ = "Sam Schott"
__url__ = "https://github.com/SamSchott/maestral"
